########################################################################
# Introscope EPAgent Process Perl Library
#                                                                      
# Wily Technology (tm) Introscope(R) Version 4 Release 4.2
# Copyright (c) 1998 - 2003 Wily Technology, Inc. All Rights Reserved.
# Introscope(R) is a registered trademark of Wily Technology, Inc.
#
########################################################################
# Process Module
########################################################################

package Wily::Process;

use strict;

#----------------
# Import Modules
#----------------
use FindBin;
use lib ("$FindBin::Bin", "$FindBin::Bin/.");
use Wily::PrintMetric;
use Getopt::Long;
use Carp;

# Define Sub-routines
#
sub reportDeadProcess($\%\%) {
	my $resource = shift();
	my %process = %{ shift() };
	my %process_data = %{ shift() };
	### print "### reportDeadProcess: Application $process{UNIQUE_ID} no longer exists!\n";
	Wily::PrintMetric::printMetric( type        => $process_data{ALIVE}{datatype},
					resource    => $resource,
					subresource => $process{UNIQUE_ID},
					name        => $process_data{ALIVE}{description},
					value       => 0,
	);
	if ( defined $process{PID} ) {
		Wily::PrintMetric::printMetric( type        => $process_data{PID}{datatype},
						resource    => $resource,
						subresource => $process{UNIQUE_ID},
						name        => $process_data{PID}{description},
						value       => $process{PID},
		);
	}
	# reset some numerical values
	while ((my $key,my $value) = each %process) {
		if ( defined $process_data{ $key } && $process_data{ $key }{datatype} =~ /Counter/i ) { 
			Wily::PrintMetric::printMetric( type        => $process_data{ $key }{datatype},
							resource    => $resource,
							subresource => $process{UNIQUE_ID},
							name        => $process_data{ $key }{description},
							value       => 0,
			);
		}
	}
}

sub reportAliveProcess($\%\%) {
	my $resource = shift();
	my %process = %{ shift() };
	my %process_data = %{ shift() };
	### print "### reportAliveProcess: Application $process{UNIQUE_ID} is alive!\n";
	Wily::PrintMetric::printMetric( type        => $process_data{ALIVE}{datatype},
					resource    => $resource,
					subresource => $process{UNIQUE_ID},
					name        => $process_data{ALIVE}{description},
					value       => 1,
	);
	while ((my $key,my $value) = each %process) {
		if ( defined $value && defined $process_data{ $key } ) {
			### print "### reportAliveProcess: $key=$value\n";
			Wily::PrintMetric::printMetric( type        => $process_data{ $key }{datatype},
							resource    => $resource,
							subresource => $process{UNIQUE_ID},
							name        => $process_data{ $key }{description},
							value       => $value,
			);
		}
	}
}

sub pingProcesses {
	my $OSNAME = shift;
	my $PID = shift;
	my %processes;
	### print "### os=$OSNAME\n";
	if ( $OSNAME =~ /linux/i ) {
		%processes = &pingProcessesUnix( $PID );
	}
	elsif ( $OSNAME =~ /hpux/i || $OSNAME =~ /hp-ux/i ) {
		%processes = &pingProcessesHPUX( $PID );
	}
	elsif ( $OSNAME =~ /SunOS/i || $OSNAME =~ /Solaris/i ) {
		%processes = &pingProcessesSolaris( $PID );
	}
	elsif ( $OSNAME =~ /aix/i ) {
		%processes = &pingProcessesAIX( $PID );
	}
	elsif ( $OSNAME =~ /os390/i || $OSNAME =~ /os\/390/i ) {
		%processes = &pingProcessesUnix( $PID );
	}
	elsif ( $OSNAME =~ /unix/i ) {
		%processes = &pingProcessesUnix( $PID );
	}
	else {
		die "Platform not supported or tested: $OSNAME\nIf on unsupported platform, you can try testing by using generic 'unix'.\n";
	}
}

sub pingProcessesSolaris{ 
	my $PID = shift;
	my $qualifier = '-ae';
	if ( defined $PID && $PID >= 0 ) {
		$qualifier = '-p '. $PID;
	}
	my $psCommand = '/usr/bin/ps ' . $qualifier . ' -o pid,ppid,ruser,pcpu,vsz,rss,pmem,etime,args';
	### print "### pingProcessesSolaris: ps=$psCommand\n";
	my @ps = `$psCommand | grep -v '$psCommand' | grep -v grep`;
	# Output from ps command is:
	#/home/support> /usr/bin/ps -p 1 -o pid,ppid,ruser,pcpu,vsz,rss,pmem,etime,args
	#  PID  PPID    RUSER %CPU  VSZ  RSS %MEM     ELAPSED COMMAND
	#    1     0     root  0.0 1224  360  0.1 82-22:51:06 /etc/init
	#
	# Where the following are defined:
	# - %CPU The ratio of CPU time used recently to CPU time available in the same period, expressed as a percentage. The meaning of ``recently'' in this context is unspecified. The CPU time available is determined in an unspecified manner. 
	# - VSZ The total size of the process in virtual memory, in kilobytes. 
	# - RSS The resident set size of the process, in kilobytes. 
	# - %MEM The ratio of the process's resident set size to the physical memory on the machine, expressed as a percentage. 
	# - ELAPSED In the POSIX locale, the elapsed time since the process was started, in the form: 
	#
	my %processes;
	if ( $#ps == 0 ) {
		### print "### pingProcessesSolaris:WARNING: nothing returned from ps\n";
		return %processes;
	}
	# Remove the first line which contains column headers
	@ps = @ps[1..$#ps];
	foreach my $procline (@ps) {
		chomp $procline; # get rid of trailing '\n'
		$procline =~ s/^\s*(.*?)//; # remove leading whitespace
		my @ps_data = split(" ", $procline);
		my %process;
		$process{ PID } = $ps_data[ 0 ];
		$process{ PPID } = $ps_data[ 1 ];
		$process{ USER } = $ps_data[ 2 ];
		$process{ CPU_PERCENT } = int ($ps_data[ 3 ]);
		$process{ VIRTUAL_MEM_BYTES } = $ps_data[ 4 ] * 1024;
		$process{ MEM_BYTES } = $ps_data[ 5 ] * 1024;
		$process{ MEM_PERCENT } = int($ps_data[ 6 ]);
		$process{ ELAPSED_TIME }  = $ps_data[ 7 ];
		$process{ COMMAND } = $ps_data[8];
		for ( my $i = 9; $i <= $#ps_data; $i++ ) {
			$process{ COMMAND } = $process{ COMMAND } . ' ' . $ps_data[$i];
		}
		$processes{ $process{PID} } = \%process;
	}
	return %processes;
}

sub pingProcessesAIX{
	my $PID = shift;
	my $qualifier = '-ae';
	if ( defined $PID && $PID >= 0 ) {
		$qualifier = '-p '. $PID;
	}
	my $psCommand = '/usr/bin/ps ' . $qualifier . ' -o pid,ppid,ruser,pcpu,vsz,pmem,etime,bnd,pri,stat,time,thcount,scount,args';
	### print "### pingProcessesAIX: ps=$psCommand\n";
	my @ps = `$psCommand | grep -v '$psCommand' | grep -v grep`;
	# Output from ps command is:
	# /usr/bin/ps -p 1 -o pid,ppid,ruser,pcpu,vsz,pmem,etime,bnd,pri,stat,time,thcount,scount,args
	#  PID  PPID    RUSER  %CPU   VSZ  %MEM     ELAPSED BND PRI ST        TIME THCNT SC COMMAND
	#    1     0     root   0.0  1840   0.0 81-23:05:51   0  60 A     00:03:30     1  1 /etc/init
	#
	# Where the following are defined:
	# - %CPU The ratio of CPU time used recently to CPU time available in the same period, expressed as a percentage. The meaning of ``recently'' in this context is unspecified. The CPU time available is determined in an unspecified manner. 
	# - VSZ The total size of the process in virtual memory, in kilobytes. 
	# - %MEM The ratio of the process's resident set size to the physical memory on the machine, expressed as a percentage. 
	# - ELAPSED In the POSIX locale, the elapsed time since the process was started, in the form: 
	# - BND Indicates to which (if any) processor a process or kernel thread is bound.
	# - PRI The priority of the process or kernel thread ; higher numbers mean lower priority.
	# - ST The state of the process or kernel thread 
	# - TIME The cumulative CPU time since the process started.
	# - THCNT The number of kernel threads owned by the process. 
	# - SC The suspend count for a kernel thread. 
	#
	my %processes;
	if ( $#ps == 0 ) {
		### print "### pingProcessesAIX:WARNING: nothing returned from ps\n";
		return %processes;
	}
	# Remove the first line which contains column headers
	@ps = @ps[1..$#ps];
	foreach my $procline (@ps) {
		chomp $procline; # get rid of trailing '\n'
		$procline =~ s/^\s*(.*?)//; # remove leading whitespace
		my @ps_data = split(" ", $procline);
		my %process;
		$process{ PID } = $ps_data[ 0 ];
		$process{ PPID } = $ps_data[ 1 ];
		$process{ USER } = $ps_data[ 2 ];
		$process{ CPU_PERCENT } = int($ps_data[ 3 ]);
		$process{ VIRTUAL_MEM_BYTES } = $ps_data[ 4 ] * 1024;
		$process{ MEM_PERCENT } = int($ps_data[ 5 ]);
		$process{ ELAPSED_TIME } = $ps_data[ 6 ];
		$process{ BOUND } = $ps_data[ 7 ];
		$process{ PRI } = $ps_data[ 8 ];
		$process{ STATE } = $ps_data[ 9 ];
		$process{ CPU_TIME } = $ps_data[ 10 ];
		$process{ THREADS } = $ps_data[ 11 ];
		$process{ THREADS_SUSPENDED } = $ps_data[ 12 ];
		$process{ COMMAND } = $ps_data[ 13 ];
		for ( my $i = 14; $i <= $#ps_data; $i++ ) {
			$process{ COMMAND } = $process{ COMMAND } . ' ' . $ps_data[$i];
		}
		$processes{ $process{PID} } = \%process;
	}
	return %processes;
}

# A general process monitor for UNIX
# Note: NO GUARANTEE THIS WORKS ON EVERY UNIX, BUT IT SHOULD GET MINIMAL DATA
#
sub pingProcessesUnix{ 
	my $PID = shift;
	my $qualifier = '-ae';
	if ( defined $PID && $PID >= 0 ) {
		$qualifier = '-p '. $PID;
	}
	# The -o arg might not be supported, unless the binary is XPG4
	my $psCommand = 'ps ' . $qualifier . ' -o pid,ppid,ruser,pcpu,vsz,pmem,etime,comm';
	### print "### pingProcessesUnix: ps=$psCommand\n";
	my @ps = `$psCommand | grep -v '$psCommand' | grep -v grep`;
	my %processes;
	if ( $#ps == 0 ) {
		### print "### pingProcessesUnix:WARNING: nothing returned from ps\n";
		return %processes;
	}
	# Remove the first line which contains column headers
	@ps = @ps[1..$#ps];
	foreach my $procline (@ps) {
		chomp $procline; # get rid of trailing '\n'
		$procline =~ s/^\s*(.*?)//; # remove leading whitespace
		my @ps_data = split(" ", $procline);
		my %process;
		$process{ PID } = $ps_data[ 0 ];
		$process{ PPID } = $ps_data[ 1 ];
		# On Unix platforms, there should at least be PID and PPID
		# If there's less than 8 columns returned, on UNIX that does not match
		# above referenced ps column headers requested.  So just return out and 
		# report PID and PPID
		if ( $#ps_data < 7 ) {
			### print "### pingProcessesUnix:WARNING: not enough data returned from ps: $procline\n";
			next;
		}
		$process{ USER } = $ps_data[ 2 ];
		$process{ CPU_PERCENT } = int($ps_data[ 3 ]);
		$process{ VIRTUAL_MEM_BYTES } = $ps_data[ 4 ] * 1024;
		$process{ MEM_PERCENT } = int($ps_data[ 5 ]);
		$process{ ELAPSED_TIME } = $ps_data[ 6 ];
		$process{ COMMAND } = $ps_data[ 7 ];
		for ( my $i = 8; $i <= $#ps_data; $i++ ) {
			$process{ COMMAND } = $process{ COMMAND } . ' ' . $ps_data[$i];
		}
		$processes{ $process{PID} } = \%process;
	}
	return %processes;
}

sub pingProcessesHPUX{ 
	my $PID = shift;
	my $qualifier = '-aef';
	if ( defined $PID && $PID >= 0 ) {
		### print "### pingProcessesHPUX: PID=$PID\n";
		$qualifier = '-f -p '. $PID;
	}
	my $psCommand = '/usr/bin/ps ' . $qualifier;
	### print "### pingProcessesHPUX: ps=$psCommand\n";
	my @ps = `$psCommand | grep -v '$psCommand' | grep -v grep`;
	# ps -aef
	#UID   PID  PPID  C    STIME TTY       TIME COMMAND
	#root     0     0  0  Jan  1  ?         0:10 swapper
	my %processes;
	if ( $#ps == 0 ) {
		### print "### pingProcessesHPUX:WARNING: nothing returned from ps\n";
		return %processes;
	}
	# Remove the first line which contains column headers
	@ps = @ps[1..$#ps];
	foreach my $procline (@ps) {
		chomp $procline; # get rid of trailing '\n'
		$procline =~ s/^\s*(.*?)//; # remove leading whitespace
		my @ps_data = split(" ", $procline);
		my %process;
		my $idx = 0;
		$process{ USER } = $ps_data[ $idx++ ];
		$process{ PID } = $ps_data[ $idx++ ];
		### print "### pingProcessesHPUX: PID=$process{PID}\n";
		$process{ PPID } = $ps_data[ $idx++ ];
		$process{ CPU_PERCENT } = int($ps_data[ $idx++ ]);
		$process{ START_TIME } = $ps_data[ $idx++ ];
		if ( $process{ START_TIME } =~ /^Jan/ ||
			$process{ START_TIME } =~ /^Feb/ ||
			$process{ START_TIME } =~ /^Mar/ ||
			$process{ START_TIME } =~ /^Apr/ ||
			$process{ START_TIME } =~ /^May/ ||
			$process{ START_TIME } =~ /^Jun/ ||
			$process{ START_TIME } =~ /^Jul/ ||
			$process{ START_TIME } =~ /^Aug/ ||
			$process{ START_TIME } =~ /^Sep/ ||
			$process{ START_TIME } =~ /^Oct/ ||
			$process{ START_TIME } =~ /^Nov/ ||
			$process{ START_TIME } =~ /^Dec/
		) {
			$process{ START_TIME } = $process{ START_TIME } . ' ' . $ps_data[$idx++];
		}
		# Skip next field (terminal)
		$idx++;
		$process{ CPU_TIME } = $ps_data[ $idx++ ];
		$process{ COMMAND } = $ps_data[ $idx++ ];
		for ( my $i = $idx; $i <= $#ps_data; $i++ ) {
			$process{ COMMAND } = $process{ COMMAND } . ' ' . $ps_data[$i];
		}
		$processes{ $process{PID} } = \%process;
	}
	return %processes;
}

sub defineProcessData {
# Process monitoring data fields, descriptions
# For each hash defined, add to %PROCESS_DATA hash below
#
my %ALIVE = ( description => 'Alive', datatype => 'IntCounter');
my %PID = ( description => 'Process ID', datatype => 'StringEvent', WinCounterIndex => 784);
my %COMMAND = ( description=> 'Startup Command', datatype => 'StringEvent');
my %PPID = ( description=> 'Parent process ID', datatype => 'StringEvent');
my %USER = ( description=> 'Process owner', datatype => 'StringEvent');
my %CPU_PERCENT = ( description=> 'CPU Used vs. available (%)', datatype => 'IntCounter');
my %VIRTUAL_MEM_BYTES = ( description=> 'Virtual memory (bytes)', datatype => 'LongCounter', WinCounterIndex => 186 );
# same as above
#my %PAGE_FILE_BYTES = ( description=> 'Page files (bytes)', datatype => 'LongCounter', WinCounterIndex => 184 );
my %MEM_BYTES = ( description=> 'Resident set size (bytes)', datatype => 'LongCounter', WinCounterIndex => 180 );
my %MEM_PERCENT = ( description=> 'Resident set size (%)', datatype => 'IntCounter');
my %ELAPSED_TIME = ( description=> 'Elapsed Time', datatype => 'StringEvent', WinCounterIndex => 684 );
my %START_TIME = ( description=> 'Start Time', datatype => 'StringEvent'  );
my %BOUND = ( description=> 'Bound to processor (if any)', datatype => 'StringEvent');
my %PRI = ( description=> 'Priority', datatype => 'StringEvent', WinCounterIndex => 682 );
my %STATE = ( description=> 'State', datatype => 'StringEvent');
my %CPU_TIME = ( description=> 'CPU Time', datatype => 'StringEvent' );
my %CPU_TIME_MILLIS = ( description=> 'CPU Time (ms)', datatype => 'LongCounter', WinCounterIndex => 6 );
my %PRIV_TIME_PERCENT = ( description=> 'Privileged Time (%)', datatype => 'IntCounter', WinCounterIndex => 144 );
my %USER_TIME_PERCENT = ( description=> 'User Time (%)', datatype => 'IntCounter', WinCounterIndex => 142 );
my %THREADS = ( description=> 'Kernel threads', datatype => 'IntCounter', WinCounterIndex => 680 );
my %THREADS_SUSPENDED = ( description=> 'Kernel threads suspended', datatype => 'IntCounter');
my %PAGE_FAULTS_RATE = ( description=> 'Page Faults per sec', datatype => 'IntCounter', WinCounterIndex => 28 );
my %CONTROL_OPS_RATE = ( description=> 'Control ops per sec', datatype => 'IntCounter', WinCounterIndex => 1414 );
my %WRITE_BYTES_RATE = ( description=> 'Write bytes per sec', datatype => 'LongCounter', WinCounterIndex => 1418 );
my %READ_BYTES_RATE = ( description=> 'Read bytes per sec', datatype => 'LongCounter', WinCounterIndex => 1416 );
my %WRITE_OPS_RATE = ( description=> 'Write ops per sec', datatype => 'IntCounter', WinCounterIndex => 1412 );
my %READ_OPS_RATE = ( description=> 'Read ops per sec', datatype => 'IntCounter', WinCounterIndex => 1410 );

my %PROCESS_DATA = (
	ALIVE => \%ALIVE,
	PID => \%PID,
	COMMAND => \%COMMAND,
	PPID => \%PPID,
	USER => \%USER,
	CPU_PERCENT => \%CPU_PERCENT,
	VIRTUAL_MEM_BYTES => \%VIRTUAL_MEM_BYTES,
#	PAGE_FILE_BYTES => \%PAGE_FILE_BYTES,
	MEM_BYTES => \%MEM_BYTES,
	MEM_PERCENT => \%MEM_PERCENT,
	ELAPSED_TIME => \%ELAPSED_TIME,
	START_TIME => \%START_TIME,
	BOUND => \%BOUND,
	PRI => \%PRI,
	STATE => \%STATE,
	CPU_TIME => \%CPU_TIME,
	CPU_TIME_MILLIS => \%CPU_TIME_MILLIS,
	PRIV_TIME_PERCENT => \%PRIV_TIME_PERCENT,
	USER_TIME_PERCENT => \%USER_TIME_PERCENT,
	THREADS => \%THREADS,
	THREADS_SUSPENDED => \%THREADS_SUSPENDED,
	PAGE_FAULTS_RATE => \%PAGE_FAULTS_RATE,
	CONTROL_OPS_RATE=> \%CONTROL_OPS_RATE,
	WRITE_BYTES_RATE=> \%WRITE_BYTES_RATE,
	READ_BYTES_RATE=> \%READ_BYTES_RATE,
	WRITE_OPS_RATE=> \%WRITE_OPS_RATE,
	READ_OPS_RATE=> \%READ_OPS_RATE,
);
return %PROCESS_DATA;
}

1;

