########################################################################
# Introscope EPAgent Plugin Script                               
#                                                                      
# Wily Technology (tm) Introscope(R) Version 4 Release 4.0
# Copyright (c) 1998 - 2003 Wily Technology, Inc. All Rights Reserved.
# Introscope(R) is a registered trademark of Wily Technology, Inc.
#
########################################################################

#import our modules
use FindBin;
use lib ("$FindBin::Bin", "$FindBin::Bin/../lib/perl");
use Wily::Config; # Config file parser
use Wily::PrintMetric; # log reading
use Wily::Process;
use Carp;

use strict;
use Getopt::Long;

$| = 1; # auto-flush STDOUT

# Define the process data fields and their metadata
#
my %PROCESS_DATA = Wily::Process::defineProcessData();

# load the config options from the config file
#
my $config = Wily::Config::parseConfig("$FindBin::Bin/processMonitor.cfg");
# Get and command line args which, if set, override the config file
#
GetOptions(
	'resource=s'		=> \($config->{RESOURCE}),
	'groupBy=s'		=> \($config->{GROUP_BY}),
	'platform=s'		=> \($config->{PLATFORM}),
	'match=s'		=> \($config->{MATCH}),
        'distinctMatch=s' 	=> \($config->{DISTINCT_MATCH}),
	'sleepTime=i'		=> \($config->{SLEEPTIME}),
	'n=i'			=> \($config->{NUMLOOPS}),
);

# Change this to change how many are reported in the main resource
my $TOP_N = $config->{TOP_N};

my $THIS_PID = $$;
my $OSNAME = $^O;
if ( defined $config->{PLATFORM} ) {
	$OSNAME = $config->{PLATFORM};
}
if ( $OSNAME =~ /Win/i ) {
	die "Please use the Windows version of this Perl script.";
}
if ( $config->{NUMLOOPS} < 0 ) {
	$config->{NUMLOOPS} = 0;
}
if (  $config->{SLEEPTIME} <= 0 ) {
	$config->{SLEEPTIME} = 15;
}

my $loops = 0;
my %processesByUniqueID;
my %processesByPID;
if ( defined $config->{MATCH} ) {
	if ( defined $config->{DISTINCT_MATCH} ) {
		Wily::PrintMetric::printMetric( type        => 'StringEvent',
						resource    => $config->{RESOURCE},
						name        => 'Exact Matching',
						value       => '$config->{MATCH}',
		);
	}
	else {
		Wily::PrintMetric::printMetric( type        => 'StringEvent',
						resource    => $config->{RESOURCE},
						name        => 'Matching',
						value       => '$config->{MATCH}',
		);
	}
}
while ( $config->{NUMLOOPS} == 0 || $loops < $config->{NUMLOOPS} ) {
	$loops++;
	my %processes = Wily::Process::pingProcesses $OSNAME;
	foreach my $PID (keys %processes) {
		my %process = %{$processes{ $PID }};
		# Don't report the process of this script itself
		#
		if ( $process{PID} == $THIS_PID ) {
			next;
		}
		### print "### pid = $process{PID}\n";
		if ( defined $config->{MATCH} ) {
			my $regExp = $config->{MATCH};
			### print "### matching $regExp\n";
			if ( defined $config->{DISTINCT_MATCH} && $process{COMMAND} ne $regExp ) {
				### print "### no exact match!\n";
				next 
			}
			elsif ( $process{COMMAND} !~ /$regExp/i ) {
				### print "### no match!\n";
				next;
			}
		}
		my $id;
		# Check to see if the process has already been recorded...
		# and if so, get the unique id.
		if ( exists $processesByPID{ $process{PID} } ) {
			$id = $processesByPID{ $process{PID} }{UNIQUE_ID};
		}
		elsif ( $config->{GROUP_BY} =~ /pid/i ) { # since there's no record, use the pid if grouping by pid
			$id = $process{PID};
		}
		elsif ( exists $processesByUniqueID{$process{COMMAND}} ) { # not grouping by pid, so see if there's already a process with this command
			$id = $process{COMMAND} . ' (PID=' . $process{PID} . ')';
		}
		else { # no process with this command...
			$id = $process{COMMAND};
		}
		### print "### unique id = $id\n";
		$process{UNIQUE_ID} = $id;
		# add the loops as a hashed value so we know that this process is still active
		$process{LOOPS} = $loops;
		# redefine processes hashes
		$processesByPID{ $process{PID} } = \%process;
		$processesByUniqueID{ $process{UNIQUE_ID} } = \%process;
		### print "### process loops=$process{LOOPS}\n";
	}
	foreach my $PID (keys %processesByPID) {
		my %process = %{$processesByPID{ $PID }};
		### print "### process loops=$process{LOOPS}\n";
		if ( $process{LOOPS} == $loops ) {
			Wily::Process::reportAliveProcess $config->{RESOURCE}, %process, %PROCESS_DATA;
		}
		else {
			Wily::Process::reportDeadProcess $config->{RESOURCE}, %process, %PROCESS_DATA;
			delete $processesByPID{ $PID };
			delete $processesByUniqueID{ $process{UNIQUE_ID} };
		}
	}
	my $topN;
	my $maxIdx = $TOP_N - 1;
	# Sort in reverse order to get processes with highest CPU
	my @keys = ();
	@keys = sort { $processesByPID{$b}{CPU_PERCENT} <=> $processesByPID{$a}{CPU_PERCENT} } keys %processesByPID;
	if ( $#keys < $maxIdx ) {
		$maxIdx = $#keys;
	}
	for my $key ( @keys[0..$maxIdx] ) {
		$topN = $topN . $key . "=" . $processesByPID{ $key }{CPU_PERCENT} . "%,";
	}
	if ( $#keys >= 0 ) {
		Wily::PrintMetric::printMetric( type        => 'IntCounter',
						resource    => $config->{RESOURCE},
						name        => 'Total Alive',
						value       => $#keys+1,
		);
		Wily::PrintMetric::printMetric( type        => 'StringEvent',
						resource    => $config->{RESOURCE},
						name        => 'Top ' . $TOP_N . ' by CPU',
						value       => $topN,
		);
	}
	else {
		Wily::PrintMetric::printMetric( type        => 'IntCounter',
						resource    => $config->{RESOURCE},
						name        => 'Total Alive',
						value       => 0,
		);
		Wily::PrintMetric::printMetric( type        => 'StringEvent',
						resource    => $config->{RESOURCE},
						name        => 'Top ' . $TOP_N . ' by CPU',
						value       => 'None',
		);
	}
	sleep $config->{SLEEPTIME};
}
